"""Definition of an ElkM1 Keypad."""
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, ka_encode


class Keypad(Element):
    """Class representing an Keypad"""
    def __init__(self, index):
        super().__init__(index)
        self.area = None
        self.temperature = None

class Keypads(Elements):
    """Handling for multiple areas"""
    def __init__(self, elk):
        super().__init__(elk, Keypad, Max.KEYPADS.value)
        add_message_handler('KA', self._ka_handler)
        add_message_handler('LW', self._lw_handler)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.elk.send(ka_encode())
        self.get_descriptions(TextDescriptions.KEYPAD.value)

    def _ka_handler(self, keypad_areas):
        for keypad in self.elements:
            if keypad_areas[keypad.index] >= 0:
                keypad.setattr('area', keypad_areas[keypad.index])

    # pylint: disable=unused-argument
    def _lw_handler(self, keypad_temps, zone_temps):
        for keypad in self.elements:
            if keypad_temps[keypad.index] > -40:
                keypad.setattr('temperature', keypad_temps[keypad.index])



"""Definition of an ElkM1 Task"""
from .const import Max, TextDescriptions
from .elements import Element, Elements


class Task(Element):
    """Class representing an Task"""

    def __init__(self, index): # pylint: disable=useless-super-delegation
        super().__init__(index)

class Tasks(Elements):
    """Handling for multiple tasks"""
    def __init__(self, elk):
        super().__init__(elk, Task, Max.TASKS.value)

    def sync(self):
        """Retrieve tasks from ElkM1"""
        self.get_descriptions(TextDescriptions.TASK.value)



"""Definition of an ElkM1 User"""
from .const import Max, TextDescriptions
from .elements import Element, Elements


class User(Element):
    """Class representing an User"""
    def __init__(self, index): # pylint: disable=useless-super-delegation
        super().__init__(index)

class Users(Elements):
    """Handling for multiple areas"""
    def __init__(self, elk):
        super().__init__(elk, User, Max.USERS.value)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.get_descriptions(TextDescriptions.USER.value)



"""
Constants used across package
"""

from enum import Enum

class Max(Enum):
    """Max number of elements on the panel"""
    AREAS = 8
    COUNTERS = 64
    KEYPADS = 16
    OUTPUTS = 208
    SETTINGS = 20
    TASKS = 32
    THERMOSTATS = 16
    USERS = 203
    LIGHTS = 256
    ZONES = 208
    ZONE_TEMPS = 16

class ZoneType(Enum):
    """Types of Elk zones"""
    Disabled = 0
    BurlarEntryExit1 = 1
    BurlarEntryExit2 = 2
    BurglarPerimeterInstant = 3
    BurglarInterior = 4
    BurglarInteriorFollower = 5
    BurglarInteriorNight = 6
    BurglarInteriorNightDelay = 7
    Burglar24Hour = 8
    BurglarBoxTamper = 9
    FireAlarm = 10
    FireVerified = 11
    FireSupervisory = 12
    AuxAlarm1 = 13
    AuxAlarm2 = 14
    KeyFob = 15
    NonAlarm = 16
    CarbonMonoxide = 17
    EmergencyAlarm = 18
    FreezeAlarm = 19
    GasAlarm = 20
    HeatAlarm = 21
    MedicalAlarm = 22
    PoliceAlarm = 23
    PoliceNoIndication = 24
    WaterAlarm = 25
    KeyMomentaryArmDisarm = 26
    KeyMomentaryArmAway = 27
    KeyMomentaryArmStay = 28
    KeyMomentaryDisarm = 29
    KeyOnOff = 30
    MuteAudibles = 31
    PowerSupervisory = 32
    Temperature = 33
    AnalogZone = 34
    PhoneKey = 35
    IntercomKey = 36

class ZonePhysicalStatus(Enum):
    """Zone physical status; name capitalized so can be used in displays"""
    Unconfigured = 0
    Open = 1
    EOL = 2
    Short = 3

class ZoneLogicalStatus(Enum):
    """Zone logical status; name capitalized so can be used in displays"""
    Normal = 0
    Troubled = 1
    Violated = 2
    Bypassed = 3

class TextDescriptions(Enum):
    """Types of description strings that can be retrieved from the panel"""
    ZONE = (0, Max.ZONES.value)
    AREA = (1, Max.AREAS.value)
    USER = (2, Max.USERS.value)
    KEYPAD = (3, Max.KEYPADS.value)
    OUTPUT = (4, 64)
    TASK = (5, Max.TASKS.value)
    TELEPHONE = 6
    LIGHT = (7, Max.LIGHTS.value)
    ALARM_DURATION = 8
    SETTING = (9, Max.SETTINGS.value)
    COUNTER = (10, Max.COUNTERS.value)
    THERMOSTAT = 11
    FUNCTION_KEY_1 = 12
    FUNCTION_KEY_2 = 13
    FUNCTION_KEY_3 = 14
    FUNCTION_KEY_4 = 15
    FUNCTION_KEY_5 = 16
    FUNCTION_KEY_6 = 17
    AUDIO_ZONE = 18
    AUDIO_SOURCE = 19

class ArmLevel(Enum):
    """Arming levels."""
    Disarm = '0'
    ArmAway = '1'
    ArmStay = '2'
    ArmStayInstant = '3'
    ArmNight = '4'
    ArmNightInstant = '5'
    ArmVacation = '6'
    ArmStepToNextAwayMode = '7'
    ArmStepToNextStayMode = '8'
    ForceArmAway = '9'
    ForceArmStay = ':'

# Map to convert message code to descriptive string
MESSAGE_MAP = {
    'AP': "Send ASCII String",
    'AR': "Alarm Reporting to Ethernet",
    'AS': "Arming status report data",
    'AT': "Ethernet Test to IP",
    'AZ': "Alarm by zone reply",
    'CA': "Reply Touchscreen audio command",
    'CC': "Control output change update",
    'CD': "Outgoing Audio Equip Command",
    'CR': "Custom value report data",
    'CS': "Control output status report data",
    'CU': "Change user code reply",
    'CV': "Counter Value Data",
    'DK': "Display KP LCD Data, not used",
    'DS': "Lighting Poll Response",
    'EE': "Entry/Exit Time Data",
    'EM': "Email Trigger to M1XEP",
    'IC': "Send invalid user code digits",
    'IE': "Installer program exited",
    'IP': "M1XSP Insteon Program",
    'IR': "M1XSP Insteon Read",
    'KA': "Keypad areas report data",
    'KC': "Keypad key change update",
    'KF': "Function key pressed data",
    'LD': "Log data with index",
    'LW': "Reply temperature data",
    'NS': "Reply Source Name",
    'NZ': "Reply Zone Name",
    'PC': "PLC change update",
    'PS': "PLC status report data",
    'RE': "Reset Ethernet Module",
    'RP': "ELKRP connected",
    'RR': "Real Time Clock Data",
    'SD': "Text string description report data",
    'SS': "System Trouble Status data",
    'ST': "Temperature report data",
    'T2': "Reply Omnistat 2 data",
    'TC': "Task change update",
    'TR': "Thermostat data report",
    'UA': "User code areas report data",
    'VN': "Reply Version Number of M1",
    'XB': "reserved by ELKRP",
    'XK': "Request Ethernet test",
    'ZB': "Zone bypass report data",
    'ZC': "Zone change update",
    'ZD': "Zone definition report data",
    'ZP': "Zone partition report data",
    'ZS': "Zone status report data",
    'ZV': "Zone analog voltage data",

    'a0': "Disarm",
    'a1': "Arm to away",
    'a2': "Arm to stay",
    'a3': "Arm to stay instant",
    'a4': "Arm to night",
    'a5': "Arm to night instant",
    'a6': "Arm to vacation",
    'a7': "Arm, step to next Away Mode",
    'a8': "Arm, step to next Stay Mode",
    'a9': "Force Arm to Away Mode",
    'a:': "Force Arm to Stay Mode",
    'ar': "Alarm Reporting Acknowledge",
    'as': "Request arming status",
    'at': "Ethernet Test Acknowledge",
    'az': "Alarm by zone request",
    'ca': "Request Touchscreen audio command",
    'cd': "Incoming Audio Equip Command",
    'cf': "Control output OFF",
    'cn': "Control output ON",
    'cp': "Request ALL custom values",
    'cr': "Request custom value",
    'cs': "Control output status request",
    'ct': "Control output TOGGLE",
    'cu': "Change user code request",
    'cv': "Request Counter value",
    'cw': "Write custom value data",
    'cx': "Write counter value",
    'dm': "Display message",
    'ds': "Lighting Poll Request",
    'ip': "M1XSP Insteon Program",
    'ir': "M1XSP Insteon Read",
    'ka': "Request keypad areas",
    'kc': "Request F Key illumination status",
    'kf': "Request simulated function key press",
    'ld': "Request log data, with index",
    'le': "Write Log Data Entry",
    'lw': "Request temperature data",
    'pc': "Control any PLC device",
    'pf': "Turn OFF PLC device",
    'pn': "Turn ON PLC device",
    'ps': "Request PLC status",
    'pt': "Toggle PLC device",
    'rr': "Request Real Time Clock Read",
    'rs': "Used by Touchscreen",
    'rw': "Real Time Clock Write",
    'sd': "Request text string descriptions",
    'sp': "Speak phrase",
    'ss': "Request System Trouble Status",
    'st': "Request temperature",
    'sw': "Speak word",
    't2': "Request Omnistat 2 data",
    'tn': "Task activation",
    'tr': "Request thermostat data",
    'ts': "Set thermostat data",
    'ua': "Request user code areas",
    'vn': "request Version Number of M1",
    'xk': "Reply from Ethernet test",
    'zb': "Zone bypass request",
    'zd': "Request zone definition data",
    'zp': "Zone partition request",
    'zs': "Zone status request",
    'zv': "Request Zone analog voltage",
}



import pytest
from unittest.mock import Mock, call
from elkm1.proto import Connection

def test_receive_single_packet_and_pass_to_handler():
    mock_gotdata = Mock()
    connection = Connection(gotdata_callbk=mock_gotdata)
    connection.data_received('test\r\n'.encode())
    mock_gotdata.assert_called_once_with('test')

def test_receive_multiple_packets():
    mock_gotdata = Mock()
    connection = Connection(gotdata_callbk=mock_gotdata)
    connection.data_received('test1\r\ntest2\r\n'.encode())
    mock_gotdata.assert_has_calls([call('test1'), call('test2')])

def test_received_full_packet_and_partial_packet():
    mock_gotdata = Mock()
    connection = Connection(gotdata_callbk=mock_gotdata)
    connection.data_received('test1\r\nfirst-bit,'.encode())
    mock_gotdata.assert_called_once_with('test1')

    connection.data_received('second-bit\r\n'.encode())
    mock_gotdata.assert_has_calls([call('test1'), call('first-bit,second-bit')])

def test_write_simple():
    connection = Connection()
    connection._transport = Mock()
    connection.write_data('08cv0900')
    connection._transport.write.assert_called_once_with('08cv0900F6\r\n'.encode())



"""Definition of an ElkM1 Light"""

from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, ps_encode


class Light(Element):
    """Class representing a Light"""
    def __init__(self, index):
        super().__init__(index)
        self.status = 0

class Lights(Elements):
    """Handling for multiple lights"""
    def __init__(self, elk):
        super().__init__(elk, Light, Max.LIGHTS.value)
        add_message_handler('PC', self._pc_handler)
        add_message_handler('PS', self._ps_handler)

    def sync(self):
        """Retrieve lights from ElkM1"""
        for i in range(4):
            self.elk.send(ps_encode(i))
        self.get_descriptions(TextDescriptions.LIGHT.value)

    # pylint: disable=unused-argument
    def _pc_handler(self, housecode, index, light_level):
        self.elements[index].setattr('status', light_level)

    def _ps_handler(self, bank, statuses):
        for i in range(bank*64, (bank+1)*64):
            self.elements[i].setattr('status', statuses[i-bank*64])



"""Definition of an ElkM1 Area"""
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, as_encode


class Area(Element):
    """Class representing an Area"""
    def __init__(self, index):
        super().__init__(index)
        self.armed_status = None
        self.arm_up_state = None
        self.alarm_state = None

class Areas(Elements):
    """Handling for multiple areas"""
    def __init__(self, elk):
        super().__init__(elk, Area, Max.AREAS.value)
        add_message_handler('AS', self._as_handler)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.elk.send(as_encode())
        self.get_descriptions(TextDescriptions.AREA.value)

    def _as_handler(self, armed_statuses, arm_up_states, alarm_states):
        for area in self.elements:
            area.setattr('armed_status', armed_statuses[area.index])
            area.setattr('arm_up_state', arm_up_states[area.index])
            area.setattr('alarm_state', alarm_states[area.index])



"""
  Base of all the elements found on the Elk panel... Zone, Keypad, etc.
"""

from abc import abstractmethod
from .util import add_sync_handler, get_descriptions


class Element:
    """Element class"""
    def __init__(self, index):
        self._index = index
        self._callbacks = []
        self.name = self.default_name()

    @property
    def index(self):
        """Get the index, immutable once class created"""
        return self._index

    def add_callback(self, callback):
        """Callbacks when attribute of element changes"""
        self._callbacks.append(callback)

    def remove_callback(self, callback):
        """Callbacks when attribute of element changes"""
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def _call_callbacks(self, changed_attribute, new_value):
        """Callbacks when attribute of element changes"""
        for callback in self._callbacks:
            callback(changed_attribute, new_value)

    def setattr(self, attr, new_value):
        """If attribute value has changed then set it and call the callbacks"""
        existing_value = getattr(self, attr, None)
        if existing_value == new_value:
            return
        setattr(self, attr, new_value)
        self._call_callbacks(attr, new_value)

    def default_name(self):
        """Return a default name for based on class and index of element"""
        return self.__class__.__name__ + '-{:03d}'.format(self._index)

    def is_default_name(self):
        return self.name == self.default_name()

    def __str__(self):
        varlist = {k:v for (k, v) in vars(self).items() \
           if not k.startswith('_') and k != 'name'}.items()
        varstr = ' '.join("%s:%s" % item for item in varlist)
        return "{} '{}' {}".format(self._index, self.name, varstr)

class Elements:
    """Base for list of elements."""
    def __init__(self, elk, class_, max_elements):
        self.elk = elk
        self.max_elements = max_elements
        self.elements = [class_(i) for i in range(max_elements)]
        add_sync_handler(self.sync)

    def __iter__(self):
        for element in self.elements:
            yield element

    def __getitem__(self, key):
        return self.elements[key]

    def _got_desc(self, descriptions):
        for element in self.elements:
            if element.index >= len(descriptions):
                break
            if descriptions[element.index] is not None:
                element.setattr('name', descriptions[element.index])

    @abstractmethod
    def sync(self):
        """Synchronize elements"""
        pass

    def get_descriptions(self, description_type):
        """Get the list of descriptions for the element."""
        get_descriptions(self.elk, description_type, self._got_desc)



"""Master class that combines all ElkM1 pieces together."""

import asyncio
import logging
from importlib import import_module

from .message import message_decode
from .proto import Connection
from .util import call_sync_handlers, parse_url, url_scheme_is_secure


LOG = logging.getLogger(__name__)

class Elk:
    """Represents all the components on an Elk panel."""

    def __init__(self, config, loop=None):
        """Initialize a new Elk instance."""
        self.loop = loop if loop else asyncio.get_event_loop()
        self._config = config
        self._conn = None

        # Setup for all the types of elements tracked
        if 'element_list' in config:
            element_list = config['element_list']
        else:
            element_list = ['panel', 'zones', 'lights', 'areas', 'tasks', 'keypads',
                            'outputs', 'counters', 'settings']
        for element in element_list:
            self._create_element(element)

    def _create_element(self, element):
        module = import_module('elkm1.'+element)
        class_ = getattr(module, element.capitalize())
        setattr(self, element, class_(self))

    async def connect(self):
        """Asyncio connection to Elk."""
        url = self._config['url']
        LOG.debug("Elk connect to %s", url)
        _scheme, host, port, ssl_context = parse_url(url)
        # pylint: disable=C0330
        _coro = await self.loop.create_connection(
            lambda: Connection(loop=self.loop,
                connection_made=self._connection_made,
                gotdata_callbk=self._got_data),
                host=host, port=port, ssl=ssl_context)

    def run(self):
        """Enter the asyncio loop."""
        self.loop.run_forever()

    def _got_data(self, data): # pylint: disable=no-self-use
        LOG.debug("got_data '%s'", data)
        try:
            message_decode(data)
        except ValueError as err:
            LOG.debug(err)

    def send(self, msg):
        """Send a message to Elk panel."""
        self._conn.write_data(msg.message, msg.response_command)

    def _connection_made(self, _transport, conn):
        """Login and sync the ElkM1 panel to memory."""
        self._conn = conn
        if url_scheme_is_secure(self._config['url']):
            self._conn.write_data(self._config['userid'], raw=True)
            self._conn.write_data(self._config['password'], raw=True)
        call_sync_handlers()



import pytest
from unittest.mock import Mock

from elkm1.const import TextDescriptions
from elkm1.message import message_decode, MessageEncode
from elkm1.util import parse_url, get_descriptions, sd_handler

def test_get_descriptions_sets_up_correctly():
    mock_elk = Mock()
    get_descriptions(mock_elk, TextDescriptions.ZONE.value, None)
    mock_elk.send.assert_called_with(MessageEncode('0Bsd0000100','SD'))

def test_sd_handler_get_next_works():
    mock_elk = Mock()
    mock_callback = Mock()
    desc = TextDescriptions.ZONE.value
    get_descriptions(mock_elk, desc, mock_callback)
    sd_handler(0, 0, 'Front door', False)
    mock_elk.send.assert_called_with(MessageEncode('0Bsd0000200','SD'))
    mock_callback.assert_not_called()

def test_sd_handler_ends_when_getnext_wraps():
    mock_elk = Mock()
    mock_callback = Mock()
    desc = TextDescriptions.ZONE.value
    get_descriptions(mock_elk, desc, mock_callback)
    sd_handler(0, -1, '', False)
    # print(mock_callback.method_calls)
    mock_elk.send.assert_called_with(MessageEncode('0Bsd0000100','SD'))
    mock_callback.assert_called_once_with([None] * desc[1])

def test_parse_url_valid_non_ssl_socket():
    (scheme, host, port, ssl) = parse_url('elk://some.host:1234')
    assert scheme == 'elk'
    assert host == 'some.host'
    assert port == 1234
    assert ssl is None

def test_parse_url_valid_ssl_socket():
    (scheme, host, port, ssl) = parse_url('elks://another.host:5678')
    assert scheme == 'elks'
    assert host == 'another.host'
    assert port == 5678
    assert ssl is not None

def test_parse_url_default_non_ssl_port():
    (scheme, host, port, ssl) = parse_url('elk://192.168.0.123')
    assert scheme == 'elk'
    assert host == '192.168.0.123'
    assert port == 2101
    assert ssl is None

def test_parse_url_default_ssl_port():
    (scheme, host, port, ssl) = parse_url('elks://ssl.host')
    assert scheme == 'elks'
    assert host == 'ssl.host'
    assert port == 2601
    assert ssl is not None

def test_parse_url_serial_non_implemented():
    with pytest.raises(NotImplementedError): parse_url('serial://rest')

def test_parse_url_unknown_scheme():
    with pytest.raises(ValueError): parse_url('bad_scheme://rest')



"""Definition of an ElkM1 Zone"""
from .const import Max, TextDescriptions, ZoneType, \
                   ZoneLogicalStatus, ZonePhysicalStatus
from .elements import Element, Elements
from .message import add_message_handler, zd_encode, zp_encode, zs_encode


class Zone(Element):
    """Class representing a Zone"""
    def __init__(self, index):
        super().__init__(index)
        self.definition = 0
        self.area = 0
        self.bypassed = False
        self.logical_status = 0
        self.physical_status = 0
        self.voltage = None
        self.temperature = None

    def __str__(self):
        return ("{indx:d} '{name}' type:{typ} status:{logl}/{phys}"
                " area:{area:d} v:{volt} temp:{temp}").format(
                    name=self.name, indx=self._index,
                    typ=ZoneType(self.definition).name,
                    logl=ZoneLogicalStatus(self.logical_status).name,
                    volt=self.voltage, area=self.area, temp=self.temperature,
                    phys=ZonePhysicalStatus(self.physical_status).name)

class Zones(Elements):
    """Handling for multiple zones"""
    def __init__(self, elk):
        super().__init__(elk, Zone, Max.ZONES.value)
        add_message_handler('LW', self._lw_handler)
        add_message_handler('ZB', self._zb_handler)
        add_message_handler('ZC', self._zc_handler)
        add_message_handler('ZD', self._zd_handler)
        add_message_handler('ZP', self._zp_handler)
        add_message_handler('ZS', self._zs_handler)
        add_message_handler('ZV', self._zv_handler)

    def sync(self):
        """Retrieve zones from ElkM1"""
        self.elk.send(zd_encode())
        self.elk.send(zp_encode())
        self.elk.send(zs_encode())
        self.get_descriptions(TextDescriptions.ZONE.value)

    # pylint: disable=unused-argument
    def _lw_handler(self, keypad_temps, zone_temps):
        for i in range(16):
            zone = self.elements[i]
            if zone_temps[zone.index] > -60:
                zone.setattr('temperature', zone_temps[zone.index])

    def _zb_handler(self, zone_number, zone_bypassed):
        self.elements[zone_number].setattr('bypassed', zone_bypassed)

    def _zc_handler(self, zone_number, zone_status):
        self.elements[zone_number].setattr('logical_status', zone_status[0])
        self.elements[zone_number].setattr('physical_status', zone_status[1])

    def _zd_handler(self, zone_definitions):
        for zone in self.elements:
            zone.setattr('definition', zone_definitions[zone.index])

    def _zp_handler(self, zone_partitions):
        for zone in self.elements:
            zone.setattr('area', zone_partitions[zone.index])

    def _zs_handler(self, zone_statuses):
        for zone in self.elements:
            zone.setattr('logical_status', zone_statuses[zone.index][0])
            zone.setattr('physical_status', zone_statuses[zone.index][1])

    def _zv_handler(self, zone_number, zone_voltage):
        self.elements[zone_number].setattr('voltage', zone_voltage)



"""Definition of an ElkM1 Custom Value"""
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler


class Counter(Element):
    """Class representing an Counter"""
    def __init__(self, index):
        super().__init__(index)
        self.value = None

# pylint: disable=R0903
class Counters(Elements):
    """Handling for multiple counters"""
    def __init__(self, elk):
        super().__init__(elk, Counter, Max.COUNTERS.value)
        add_message_handler('CV', self._cv_handler)

    def sync(self):
        """Retrieve values from ElkM1 on demand"""
        self.get_descriptions(TextDescriptions.COUNTER.value)

    def _cv_handler(self, counter, value):
        countr = self.elements[counter]
        countr.value = value



"""Async IO."""

import asyncio
from functools import reduce
import logging

from .message import get_elk_command, timeout_decode


LOG = logging.getLogger(__name__)

class Connection(asyncio.Protocol):
    """asyncio Protocol with line parsing and queuing writes"""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, connection_made=None, gotdata_callbk=None, loop=None):
        self._transport = None
        self._waiting_for_response = None
        self._timeout_task = None
        self._queued_writes = []
        self._connection_made = connection_made
        self._gotdata_callbk = gotdata_callbk
        self._buffer = ''
        self.loop = loop if loop else asyncio.get_event_loop()

    def connection_made(self, transport):
        LOG.debug("connected callback")
        self._transport = transport
        if self._connection_made is not None:
            self._connection_made(transport, self)

    def connection_lost(self, exc):
        LOG.debug("disconnected callback")
        self._transport = None
        # TODO: add reconnect logic

    def _response_required_timeout(self):
        timeout_decode(self._waiting_for_response)
        self._timeout_task = None
        self._waiting_for_response = None
        self._process_write_queue()

    def data_received(self, data):
        self._buffer += data.decode('ISO-8859-1')
        while "\r\n" in self._buffer:
            line, self._buffer = self._buffer.split("\r\n", 1)
            if get_elk_command(line) == self._waiting_for_response:
                self._waiting_for_response = None
                if self._timeout_task:
                    self._timeout_task.cancel()
                    self._timeout_task = None
            self._gotdata_callbk(line)
        self._process_write_queue()

    def _process_write_queue(self):
        while self._queued_writes and not self._waiting_for_response:
            to_write = self._queued_writes.pop(0)
            self.write_data(to_write[0], to_write[1], timeout=to_write[2])

    def write_data(self, data, response_required=None, timeout=5.0, raw=False):
        """Write data on the asyncio Protocol"""
        if self._transport is None:
            return

        if self._waiting_for_response:
            LOG.debug("queueing write %s", data)
            self._queued_writes.append((data, response_required, timeout))
            return

        if response_required:
            self._waiting_for_response = response_required
            if timeout > 0:
                self._timeout_task = \
                    self.loop.call_later(timeout, self._response_required_timeout)

        if not raw:
            cksum = 256 - reduce(lambda x, y: x+y, map(ord, data)) % 256
            data = data + '{:02X}'.format(cksum)
            if int(data[0:2], 16) != len(data)-2:
                LOG.debug("message length wrong: %s", data)

        LOG.debug("write_data '%s'", data)
        self._transport.write((data + '\r\n').encode())



"""Utility functions"""

import logging
import ssl
from .message import add_message_handler, sd_encode


LOG = logging.getLogger(__name__)

# pylint: disable=invalid-name
get_descriptions_in_progress = {}
sync_handlers = []

def add_sync_handler(sync_handler):
    """Register a fn that synchronizes part of the panel."""
    sync_handlers.append(sync_handler)

def call_sync_handlers():
    """Invoke the synchronization handlers."""
    LOG.debug("Synchronizing panel...")
    for sync_handler in sync_handlers:
        sync_handler()

def get_descriptions(elk, desc, callback):
    """
    Gets the descriptions for specified type
    When complete the callback is called with a list of descriptions
    """
    desc_type = desc[0]
    max_units = desc[1]
    results = [None] * max_units
    get_descriptions_in_progress[desc_type] = (max_units, callback, results, elk)
    add_message_handler('SD', sd_handler)
    elk.send(sd_encode(desc_type=desc_type, unit=0))

# pylint: disable=unused-argument
def sd_handler(desc_type, unit, desc, show_on_keypad):
    """Text description"""
    if desc_type not in get_descriptions_in_progress:
        LOG.debug("Text description response ignored")
        return

    max_units = get_descriptions_in_progress[desc_type][0]
    results = get_descriptions_in_progress[desc_type][2]
    if unit < 0 or unit >= max_units:
        callback = get_descriptions_in_progress[desc_type][1]
        callback(results)
        del get_descriptions_in_progress[desc_type]
        return

    elk = get_descriptions_in_progress[desc_type][3]
    results[unit] = desc
    elk.send(sd_encode(desc_type=desc_type, unit=unit+1))

def url_scheme_is_secure(url):
    """Check if the URL is one that requires SSL/TLS."""
    scheme, _dest = url.split('://')
    return scheme == 'elks'

def parse_url(url):
    """Parse a Elk connection string """
    scheme, dest = url.split('://')
    if scheme == 'serial':
        # No easy way to test, so not implemented; should use serial.async
        raise NotImplementedError("Elk serial connection not implemented")

    host = None
    ssl_context = None
    if scheme == 'elk':
        host, port = dest.split(':') if ':' in dest else (dest, 2101)
    elif scheme == 'elks':
        host, port = dest.split(':') if ':' in dest else (dest, 2601)
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        ssl_context.verify_mode = ssl.CERT_NONE
    else:
        raise ValueError("Invalid scheme '%s'" % scheme)
    return (scheme, host, int(port), ssl_context)
